﻿Imports System.Text.RegularExpressions

Namespace System.ComponentModel

   ''' <summary>
   ''' trys to convert anything to string and reverse. First choice is to convert by TypeConverter. If not available, eg. for Arrays (even multidimensional), trys to convert to Base64String. Cannot convert referenceTypes when no TypeConverter available 
   ''' </summary>
   Public Structure AllConverter
      Public ReadOnly FromString As Func(Of String, Object)
      Public Shadows ReadOnly ToString As Func(Of Object, String)

      Private Shared _rgxNewlineNorm As Regex

      Shared Sub New()
         Dim nl = Environment.NewLine.ToCharArray
         _rgxNewlineNorm = New Regex("(?<!".And(nl(0), ")", nl(1)))
      End Sub

      Private Shared _StructArrayToString As Func(Of Object, String) _
         = Function(o)
              'converts value-type-Arrays to Base64String
              With DirectCast(o, Array)
                 Dim lengths(.Rank - 1) As Integer
                 For i = 0 To .Rank - 1
                    lengths(i) = .GetLength(i)
                 Next
                 Return Convert.ToBase64String(Blobber.CreateBlob(.Rank, lengths, .Self))
              End With
           End Function

      Private Shared Function StructArrayFromString( tp As Type,  base64String As String) As Object
         With New Blobber(Convert.FromBase64String(base64String))
            Dim rank = .Read(Of Integer)()
            Dim lengths(rank - 1) As Integer
            Dim lBounds(rank - 1) As Integer 'all LBounds are 0
            .ReadTo(lengths)
            StructArrayFromString = Array.CreateInstance(tp, lengths, lBounds)
            .ReadTo(tp, DirectCast(StructArrayFromString, Array))
         End With
      End Function

      Private Shared _ToBase64String As Func(Of Object, String) _
         = Function(o) Convert.ToBase64String(DirectCast(o, Byte()))

      Public Sub New( tp As Type)
         Dim conv = TypeDescriptor.GetConverter(tp)
         If TypeOf conv Is StringConverter Then
            'no conversion
            FromString = Function(s) s

            'ToString = Function(o)
            '              Dim mts = _rgxNewlineNorm.Matches(DirectCast(o, String))
            '              If mts.Count > 0 Then Stop
            '              Return _rgxNewlineNorm.Replace(DirectCast(o, String), Environment.NewLine)
            '           End Function

            ToString = Function(o) DirectCast(o, String)

         ElseIf conv.CanConvertFrom(GetType(String)) Then
            'standard-conversion
            FromString = AddressOf conv.ConvertFromInvariantString
            ToString = AddressOf conv.ConvertToInvariantString
         ElseIf conv.CanConvertFrom(GetType(Byte())) Then
            'Blob-conversion: object <-> Byte() <-> Base64String 
            FromString = Function(s) conv.ConvertFrom(Convert.FromBase64String(s))
            ToString = Function(o) _ToBase64String(conv.ConvertTo(o, GetType(Byte())))
         ElseIf tp.IsValueType Then
            'Blob-conversion: object <-> Byte() <-> Base64String 
            FromString = Function(s) (New Blobber(Convert.FromBase64String(s))).Read(tp)
            ToString = Function(o) Convert.ToBase64String(Blobber.CreateBlob(o))
         ElseIf TypeOf conv Is ArrayConverter Then
            If tp Is GetType(Byte()) Then
               'convert Byte() <-> Base64String 
               FromString = AddressOf Convert.FromBase64String
               ToString = _ToBase64String
            ElseIf tp.GetElementType.IsValueType Then
               'Blob-conversion: array <-> Byte() <-> Base64String 
               FromString = Function(s) StructArrayFromString(tp.GetElementType, s)
               ToString = _StructArrayToString
            End If
         End If
         If FromString.Null Then Throw Me.Exception("can't convert that")
      End Sub
   End Structure

End Namespace
